# Copyright (c) 2011, Intel Corporation
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of Intel Corporation nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""ACPI module."""

import bits
import struct
from cpudetect import cpulib

def parse_madt(printflag=False):
    """Parse and optionally print an MADT."""
    ACPI_MADT_ENABLED = 1
    ACPI_MADT_TYPE_LOCAL_APIC = 0
    ACPI_MADT_TYPE_LOCAL_X2APIC = 9
    procid_apicid = {}
    uid_x2apicid = {}

    # Get a buffer containing the entire MADT
    madt = bits.acpi_get_table("APIC")

    # Decode the ACPI header
    size, hdr = unpack_table_header(madt, 0, printflag)
    i = size # Get the size of the header we just decoded

    # Decode the local APIC address and APIC flags fields
    fmt_str = "<II"
    local_apic_addr, apic_flags = struct.unpack_from(fmt_str, madt, i)
    if printflag:
        print "Local APIC address = 0x%x" % local_apic_addr
        print "Multiple APIC flags = 0x%x" % apic_flags
    i += struct.calcsize(fmt_str)

    # Decode each subtable
    while i != len(madt):
        subtable_type, subtable_length = struct.unpack_from("<BB", madt, i)
        i += 2
        if subtable_type == ACPI_MADT_TYPE_LOCAL_APIC:
            procid, apicid, flags = struct.unpack_from("<BBI", madt, i)
            if flags & ACPI_MADT_ENABLED:
                procid_apicid[procid] = apicid
        if subtable_type == ACPI_MADT_TYPE_LOCAL_X2APIC:
            reserved, x2apicid, flags, uid = struct.unpack_from("<HIII", madt, i)
            if flags & ACPI_MADT_ENABLED:
                uid_x2apicid[uid] = x2apicid
        i += subtable_length - 2
    if printflag:
        print procid_apicid
        print uid_x2apicid
    return (procid_apicid, uid_x2apicid)

def unpack_table_header(tbl, offset=0, printflag=False):
    """Parse, optionally print and return an unpacked ACPI header"""
    fmt_str = "<4sIBB6s8sI4sI"
    hdrsize = struct.calcsize(fmt_str)
    signature, length, revision, checksum, oemid, oemtableid, oemrevision, creatorid, creatorrevision = struct.unpack_from(fmt_str, tbl, offset)
    if printflag:
        print "Signature = %4s" % signature
        print "Length = %u" % length
        print "Revision = %u" % revision
        print "Checksum = 0x%02x" % checksum
        print "OEMID = %6s" % oemid
        print "OEM Table ID = %8s" % oemtableid
        print "OEM Revision = 0x%04x" % oemrevision
        print "Creator ID = %s" % creatorid
        print "Creator Revision = 0x%04x" % creatorrevision
    return (hdrsize, (signature, length, revision, checksum, oemid, oemtableid, oemrevision, creatorid, creatorrevision))

def find_procid(printflag=False):
    cpupaths = bits.acpi_cpupaths()
    cpupath_procid = {}
    for cpupath in cpupaths:
        value = bits.acpi_eval(cpupath)
        cpupath_procid[cpupath] = value[0]
    if printflag:
        print cpupath_procid
    return cpupath_procid

def find_uid(printflag=False):
    cpupaths = bits.acpi_cpupaths()
    cpupath_uid = {}
    for cpupath in cpupaths:
        value = bits.acpi_eval(cpupath + "._UID")
        cpupath_uid[cpupath] = value
    if printflag:
        print cpupath_uid
    return cpupath_uid

def commonprefix(l):
    """Return the common prefix of a list of strings."""
    if not l:
        return ''
    prefix = l[0]
    for s in l[1:]:
        for i, c in enumerate(prefix):
            if c != s[i]:
                prefix = s[:i]
                break
    return prefix

def factor_commonprefix(l):
    prefix = commonprefix(l)
    prefixlen = len(prefix)
    return prefix + "{" + ", ".join([s[prefixlen:] for s in l]) + "}"

def display_cpupaths():
    """Find and display ACPI processor name space paths"""
    cpupaths = bits.acpi_cpupaths()
    print factor_commonprefix(cpupaths)

def display_cpu_info():
    cpupaths = bits.acpi_cpupaths()
    cpupath_procid = find_procid()
    cpupath_uid = find_uid()
    procid_apicid, uid_x2apicid = parse_madt()
    for cpupath in cpupaths:
        print cpupath,
        def socket_str(apicid):
            socket_index = bits.socket_index(apicid)
            if socket_index is None:
                return ''
            return ', socket index = 0x%02x' % socket_index
        def apicid_str(apicid):
            if apicid is None:
                return 'no APIC ID'
            return 'APIC ID 0x%02x%s' % (apicid, socket_str(apicid))
        procid = cpupath_procid.get(cpupath, None)
        if procid is not None:
            print 'ProcID = %-2u (%s)' % (procid, apicid_str(procid_apicid.get(procid, None))),
        uid = cpupath_uid.get(cpupath, None)
        if uid is not None:
            print '_UID = %s (%s)' % (uid, apicid_str(uid_x2apicid.get(uid, None))),
        print

def display_acpi_method(method, print_one):
    """Helper function that performs all basic processing for evaluating an ACPI method"""
    cpupaths = bits.acpi_cpupaths()
    uniques = {}
    for cpupath in cpupaths:
        value = bits.acpi_eval(cpupath + "." + method)
        uniques.setdefault(value, []).append(cpupath)

    print "%u unique %s values" % (len(uniques), method)
    for value, cpupaths in sorted(uniques.iteritems(), key=(lambda (k,v): v)):
        print
        print factor_commonprefix(cpupaths)
        if value is None:
            print "No %s found for these CPUs" % method
        else:
            print_one(value)

def unpack_register(reg):
    """Unpack a register string to (AddressSpaceId, BitWidth, BitOffset, AccessSize, Address)"""
    (DescriptorType, ResourceLength, AddressSpaceId, BitWidth, BitOffset, AccessSize, Address, EndTagByte, Checksum) = struct.unpack("<BHBBBBQBB", reg)
    return (AddressSpaceId, BitWidth, BitOffset, AccessSize, Address)

def display_csd():
    """Find and display _CSD (C-State Dependency) methods"""
    def print_csd(csd):
        if not isinstance(csd, tuple):
            print "No _CSD found for these CPUs"
            return
        fmt_str = "              %-10s %-10s %-8s %-10s %-9s %-10s"
        print fmt_str % ("NumEntries", "Revision", "Domain", "CoordType", "NumProcs", "Index")
        for csd_num, [NumEntries, Revision, Domain, CoordType, NumProcessors, Index] in enumerate(csd):
            fmt_str = "Dependency[%u] 0x%-8x 0x%-8x 0x%-6x 0x%-8x 0x%-7x %-8x"
            print fmt_str % (csd_num, NumEntries, Revision, Domain, CoordType, NumProcessors, Index)
    display_acpi_method("_CSD", print_csd)

def display_cst():
    """Find and display _CST (C-States) methods"""
    def print_cst(cst):
        for (reg, cst_type, latency, power) in cst[1:]:
            (AddressSpaceId, BitWidth, BitOffset, AccessSize, Address) = unpack_register(reg)
            print
            if AddressSpaceId == 0x7f:
                # Decode register as FFH
                print "Fixed Functional Hardware (FFH) detected, using Intel-specific decode"
                Type, VendorCode, ClassCode, Arg1, Arg0 = AddressSpaceId, BitWidth, BitOffset, AccessSize, Address
                fmt_str = "%-5s %-10s %-9s %-10s %-8s"
                print fmt_str % ("Type", "VendorCode", "ClassCode", "Arg0", "Arg1")
                fmt_str = "0x%-3x 0x%-8x 0x%-7x 0x%-8x 0x%-8x"
                print fmt_str % (Type, VendorCode, ClassCode, Arg0, Arg1)
                if VendorCode == 0:
                    print "  C1 Halt"
                elif (VendorCode == 1) and (ClassCode == 1):
                    print "  C1 I/O then Halt I/O port address = 0x%x" % Arg0
                elif (VendorCode == 1) and (ClassCode == 2):
                    fmt_str = "  Native C-state instruction, Hint: 0x%x (%s)"
                    print fmt_str % (Arg0, cpulib.mwait_hint_to_cstate(Arg0))
                    fmt_str = "  HW coordination=%u, Bus Master avoidance=%u"
                    print fmt_str % (bool(Arg1 & 1), bool((Arg1 & (1 << 1))))
            else:
                # Decode register as actual hardware resource
                fmt_str = "%-11s %-10s %-9s %-10s %-8s"
                print fmt_str % ("AddrSpaceId", "BitWidth", "BitOffset", "AccessSize", "Address")
                fmt_str = "0x%-9x 0x%-8x 0x%-7x 0x%-8x 0x%-8x"
                print fmt_str % (AddressSpaceId, BitWidth, BitOffset, AccessSize, Address)
            # Decode and print ACPI c-state, latency, & power
            fmt_str = "%-7s %-7s %-7s"
            print fmt_str % ("C-state", "Latency", "Power")
            cst_type_str = "ACPI C%u" % cst_type
            latency_str = "%uus" % latency
            power_str = "%umW" % power
            print fmt_str % (cst_type_str, latency_str, power_str)
    display_acpi_method("_CST", print_cst)

def display_pct():
    """Find and display _PCT (Performance Control) methods"""
    def print_pct(pct):
        fmt_str = "%-12s %-5s %-10s %-9s %-10s %-8s"
        print fmt_str % ("", "Type",        "VendorCode","ClassCode", "Arg0",       "Arg1")
        reg_type = ("Control", "Status")
        for reg_num, reg in enumerate(pct):
            (AddressSpaceId, BitWidth, BitOffset, AccessSize, Address) = unpack_register(reg)
            fmt_str = "%-12s 0x%-3x 0x%-8x 0x%-7x 0x%-8x 0x%-8x"
            print  fmt_str % (reg_type[reg_num], AddressSpaceId, BitWidth, BitOffset, AccessSize, Address)
    display_acpi_method("_PCT", print_pct)

def display_pdl():
    """Find and display _PDL (P-State Depth Limit) methods"""
    def print_pdl(pdl):
        print "Lowest performance state that OSPM can use = %u" % pdl
    display_acpi_method("_PDL", print_pdl)

def display_ppc():
    """Find and display _PPC (Performance Present Capabilities) methods"""
    def print_ppc(ppc):
        print "Highest performance state that OSPM can use = %u" % ppc
    display_acpi_method("_PPC", print_ppc)

def display_psd():
    """Find and display _PSD (P-State Dependency) methods"""
    def print_psd(psd):
        # Verify the ACPI data returned is a package and not a NULL
        if not isinstance(psd, tuple):
            print "No _PSD found for these CPUs"
            return
        fmt_str = "               %-10s %-10s %-10s %-10s %-10s"
        print fmt_str % ("NumEntries", "Revision", "Domain", "CoordType", "NumProcessors")
        for psd_num, [NumEntries, Revision, Domain, CoordType, NumProcessors] in enumerate(psd):
            fmt_str = "Dependency[%u]: 0x%-8x 0x%-8x 0x%-8x 0x%-8x 0x%-8x"
            print fmt_str % (psd_num, NumEntries, Revision, Domain, CoordType, NumProcessors)
    display_acpi_method("_PSD", print_psd)

def display_pss():
    """Find and display _PSS (Performance Supported States) methods"""
    def print_pss(pss):
        fmt_str = "    %-10s %-10s %-10s %-10s %-10s %-10s"
        if len(pss) == 0:
            print "Empty _PSS was returned."
            return
        print fmt_str % ("Frequency", "Power", "Latency", "BMLatency", "Control", "Status")
        for pstate_num, [freq, power, latency, bmlatency, control, status] in enumerate(pss):
            fmt_str = "P%-2u 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x"
            print fmt_str % (pstate_num, freq, power, latency, bmlatency, control, status)
    display_acpi_method("_PSS", print_pss)

def display_ptc():
    """Find and display _PTC (Processor Throttling Control) methods"""
    def print_ptc(ptc):
        fmt_str = "%-12s %-13s %-12s %-11s %-12s %s"
        print fmt_str % ("Processor",  "AddrSpaceId", "BitWidth",   "BitOffset", "AccessSize", "Address")
        print fmt_str % ("Throttling", "Type",        "VendorCode", "ClassCode", "Arg0",       "Arg1")
        reg_type = ("Control", "Status")
        for reg_num, reg in enumerate(ptc):
            (AddressSpaceId, BitWidth, BitOffset, AccessSize, Address) = unpack_register(reg)
            fmt_str = "%-12s 0x%-11x 0x%-10x 0x%-9x 0x%-10x 0x%x"
            print fmt_str % (reg_type[reg_num], AddressSpaceId, BitWidth, BitOffset, AccessSize, Address)
    display_acpi_method("_PTC", print_ptc)

def display_tdl():
    """Find and display _TDL (T-State Depth Limit) methods"""
    def print_tdl(tdl):
        print "Lowest power throttling state that OSPM may use = %u" % tdl
    display_acpi_method("_TDL", print_tdl)

def display_tpc():
    """Find and display _TPC (Throttling Present Capabilities) methods"""
    def print_tpc(tpc):
        print "Highest throttling state that OSPM can use = %u" % tpc
    display_acpi_method("_TPC", print_tpc)

def display_tsd():
    """Find and display _TSD (T-State Dependency) methods"""
    def print_tsd(tsd):
        # Verify the ACPI data returned is a package and not a NULL
        if not isinstance(tsd, tuple):
            print "No _TSD found for these CPUs"
            return
        fmt_str = "               %-10s %-10s %-10s %-10s    %-10s"
        print fmt_str % ("NumEntries", "Revision", "Domain", "CoordType", "NumProcessors")
        for tsd_num, [NumEntries, Revision, Domain, CoordType, NumProcessors] in enumerate(tsd):
            fmt_str = "Dependency[%u]: 0x%-8u 0x%-8u 0x%-8x 0x%-8x    %-8u"
            print fmt_str % (tsd_num, NumEntries, Revision, Domain, CoordType, NumProcessors)
    display_acpi_method("_TSD", print_tsd)

def display_tss():
    """Find and display _TSS (Throttling Supported States) methods"""
    def print_tss(tss):
        # Verify the ACPI data returned is a package and not a NULL
        if not isinstance(tss, tuple):
            print "No _TSS found for these CPUs"
            return
        fmt_str = "    %-9s %-10s %-10s %-10s %-10s"
        print fmt_str % ("Percent", "Power", "Latency", "Control", "Status")
        for tstate_num, [percent, power, latency, control, status] in enumerate(tss):
            fmt_str = "T%-2u %6u%%   0x%08x 0x%08x 0x%08x 0x%08x"
            print fmt_str % (tstate_num, percent, power, latency, control, status)
    display_acpi_method("_TSS", print_tss)

def display_uid():
    """Find and display _UID"""
    def print_uid(uid):
        print "_UID = %s" % uid
    display_acpi_method("_UID", print_uid)
