#include "Python.h"

#include "acpica.h"
#include "smp.h"
#include "pci.h"

struct dword_regs {
    U32 eax;
    U32 ebx;
    U32 ecx;
    U32 edx;
};

static U32 pcie_base = 0;

static void cpuid_callback(void *param)
{
    struct dword_regs *r = param;
    cpuid32_indexed(r->eax, r->ecx, &r->eax, &r->ebx, &r->ecx, &r->edx);
}

static PyObject *bits_cpuid(PyObject *self, PyObject *args)
{
    struct dword_regs regs = { .ecx=0 };
    unsigned apicid;
    int ncpus;

    if (!PyArg_ParseTuple(args, "II|I", &apicid, &regs.eax, &regs.ecx))
        return NULL;

    ncpus = smp_init();
    if (!ncpus)
        return PyErr_Format(PyExc_RuntimeError, "SMP module failed to initialize.");

    if (!smp_function(apicid, cpuid_callback, &regs))
        return Py_BuildValue("");
    return Py_BuildValue("IIII", regs.eax, regs.ebx, regs.ecx, regs.edx);
}

static PyObject *bits_cpus(PyObject *self, PyObject *args)
{
    int ncpus;
    int ndx;
    PyObject *apicid_list;
    const CPU_INFO *cpu;

    ncpus = smp_init();
    if (!ncpus)
        return PyErr_Format(PyExc_RuntimeError, "SMP module failed to initialize.");

    apicid_list = PyList_New(ncpus);
    cpu = smp_read_cpu_list();
    for (ndx = 0; ndx < ncpus; ndx++)
        PyList_SET_ITEM(apicid_list, ndx, Py_BuildValue("i", cpu[ndx].apicid));
    return Py_BuildValue("O", apicid_list);
}

struct msr {
    U32 num;
    U32 status;
    U64 value;
};

static void rdmsr_callback(void *param)
{
    struct msr *m = param;
    rdmsr64(m->num, &m->value, &m->status);
}

static PyObject *bits_rdmsr(PyObject *self, PyObject *args)
{
    struct msr msr;
    unsigned apicid;
    int ncpus;

    if (!PyArg_ParseTuple(args, "II", &apicid, &msr.num))
        return NULL;

    ncpus = smp_init();
    if (!ncpus)
        return PyErr_Format(PyExc_RuntimeError, "SMP module failed to initialize.");

    if (!smp_function(apicid, rdmsr_callback, &msr))
        return Py_BuildValue("");
    if (msr.status)
        return Py_BuildValue("");
    return Py_BuildValue("K", msr.value);
}

static void wrmsr_callback(void *param)
{
    struct msr *m = param;
    wrmsr64(m->num, m->value, &m->status);
}

static PyObject *bits_wrmsr(PyObject *self, PyObject *args)
{
    struct msr msr;
    unsigned apicid;
    int ncpus;

    if (!PyArg_ParseTuple(args, "IIK", &apicid, &msr.num, &msr.value))
        return NULL;

    ncpus = smp_init();
    if (!ncpus)
        return PyErr_Format(PyExc_RuntimeError, "SMP module failed to initialize.");

    if (!smp_function(apicid, wrmsr_callback, &msr))
        return Py_BuildValue("");
    return PyBool_FromLong(!msr.status);
}

static PyObject *bits_pcie_get_base(PyObject *self, PyObject *args)
{
    return Py_BuildValue("I", pcie_base);
}

static PyObject *bits_pcie_set_base(PyObject *self, PyObject *args)
{
    U32 memaddr;

    if (!PyArg_ParseTuple(args, "I", &memaddr))
        return NULL;
    pcie_base = memaddr;
    return Py_BuildValue("");
}

static PyObject *bits_pcie_read(PyObject *self, PyObject *args, PyObject *keywds)
{
    U32 bus, dev, fn, reg, size = 0;
    U32 memaddr = pcie_base;
    U64 value = 0;
    static char *keywords[] = {"bus", "device", "function", "register", "bytes", "memaddr", NULL};

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "IIII|II", keywords, &bus, &dev, &fn, &reg, &size, &memaddr))
        return NULL;

    if (size == 0)
        size = (reg % 2) ? 1 :
               (reg % 4) ? 2 : 4;

    if (!memaddr)
        return PyErr_Format(PyExc_ValueError, "Invalid PCIE memory base address");

    switch (size) {
    case 1:
        value = pci_mmcfg_read_u8(memaddr, bus, dev, fn, reg);
        return Py_BuildValue("B", value);
        break;
    case 2:
        value = pci_mmcfg_read_u16(memaddr, bus, dev, fn, reg);
        return Py_BuildValue("H", value);
        break;
    case 4:
        value = pci_mmcfg_read_u32(memaddr, bus, dev, fn, reg);
        return Py_BuildValue("I", value);
        break;
    case 8:
        value = pci_mmcfg_read_u32(memaddr, bus, dev, fn, reg);
        value |= ((U64)pci_mmcfg_read_u32(memaddr, bus, dev, fn, reg+4)) << 32;
        return Py_BuildValue("K", value);
        break;
    default:
        return PyErr_Format(PyExc_ValueError, "Size must be 1, 2, 4, or 8 bytes");
    }
}

static PyObject *bits_pcie_write(PyObject *self, PyObject *args, PyObject *keywds)
{
    U32 bus, dev, fn, reg, size = 0;
    U32 memaddr = pcie_base;
    U64 value;
    static char *keywords[] = {"bus", "device", "function", "register", "value", "bytes", "memaddr", NULL};

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "IIIIK|II", keywords, &bus, &dev, &fn, &reg, &value, &size, &memaddr))
        return NULL;

    if (size == 0)
        size = (reg % 2) ? 1 :
               (reg % 4) ? 2 : 4;


    if (!memaddr)
        return PyErr_Format(PyExc_ValueError, "Invalid PCIE memory base address");

    switch (size) {
    case 1:
        pci_mmcfg_write_u8(memaddr, bus, dev, fn, reg, value);
        break;
    case 2:
        pci_mmcfg_write_u16(memaddr, bus, dev, fn, reg, value);
        break;
    case 4:
        pci_mmcfg_write_u32(memaddr, bus, dev, fn, reg, value);
        break;
    case 8:
        pci_mmcfg_write_u32(memaddr, bus, dev, fn, reg, value);
        pci_mmcfg_write_u32(memaddr, bus, dev, fn, reg+4, value >> 32);
        break;
    default:
        return PyErr_Format(PyExc_ValueError, "Size must be 1,2,4,or 8 bytes");
    }
    return Py_BuildValue("");
}

static PyObject *bits_pci_read(PyObject *self, PyObject *args, PyObject *keywds)
{
    U32 bus, dev, fn, reg, size = 0;
    U64 value = 0;
    static char *keywords[] = {"bus", "device", "function", "register", "bytes", NULL};

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "IIII|I", keywords, &bus, &dev, &fn, &reg, &size))
        return NULL;

    if (size == 0)
        size = (reg % 2) ? 1 :
               (reg % 4) ? 2 : 4;

    switch (size) {
    case 1:
        value = pci_read_u8(bus, dev, fn, reg);
        return Py_BuildValue("B", value);
        break;
    case 2:
        value = pci_read_u16(bus, dev, fn, reg);
        return Py_BuildValue("H", value);
        break;
    case 4:
        value = pci_read_u32(bus, dev, fn, reg);
        return Py_BuildValue("I", value);
        break;
    case 8:
        value = pci_read_u32(bus, dev, fn, reg);
        value |= ((U64)pci_read_u32(bus, dev, fn, reg+4)) << 32;
        return Py_BuildValue("K", value);
        break;
    default:
        return PyErr_Format(PyExc_ValueError, "Size must be 1,2,4,or 8 bytes");
    }
}

static PyObject *bits_pci_write(PyObject *self, PyObject *args, PyObject *keywds)
{
    U32 bus, dev, fn, reg, size = 0;
    U64 value;
    static char *keywords[] = {"bus", "device", "function", "register", "value", "bytes", NULL};

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "IIIIK|I", keywords, &bus, &dev, &fn, &reg, &value, &size))
        return NULL;

    if (size == 0)
        size = (reg % 2) ? 1 :
               (reg % 4) ? 2 : 4;

    switch (size) {
    case 1:
        pci_write_u8(bus, dev, fn, reg, value);
        break;
    case 2:
        pci_write_u16(bus, dev, fn, reg, value);
        break;
    case 4:
        pci_write_u32(bus, dev, fn, reg, value);
        break;
    case 8:
        pci_write_u32(bus, dev, fn, reg, value);
        pci_write_u32(bus, dev, fn, reg+4, value>>32);
        break;
    default:
        return PyErr_Format(PyExc_ValueError, "Size must be 1,2,4,or 8 bytes");
    }

    return Py_BuildValue("");
}

static PyObject *acpi_object_to_python(ACPI_OBJECT *obj)
{
    if (obj == NULL)
        return Py_BuildValue("");

    switch(obj->Type) {
    case ACPI_TYPE_INTEGER:
        return Py_BuildValue("K", obj->Integer.Value);
    case ACPI_TYPE_STRING:
        return Py_BuildValue("s#", obj->String.Pointer, obj->String.Length);
    case ACPI_TYPE_BUFFER:
        return Py_BuildValue("s#", obj->Buffer.Pointer, obj->Buffer.Length);
    case ACPI_TYPE_PACKAGE:
        {
            U32 ndx;
            PyObject *pkg = PyTuple_New(obj->Package.Count);
            for (ndx = 0; ndx < obj->Package.Count; ndx++)
                PyTuple_SET_ITEM(pkg, ndx, acpi_object_to_python(&obj->Package.Elements[ndx]));
            return Py_BuildValue("O", pkg);
        }
    case ACPI_TYPE_PROCESSOR:
        return Py_BuildValue("III", obj->Processor.ProcId, obj->Processor.PblkAddress, obj->Processor.PblkLength);
    default:
        return PyErr_Format(PyExc_RuntimeError,
                            "Unable to parse the ACPI object returned from acpi_object_to_python on unhandled ACPI_OBJECT_TYPE %u.",
                            obj->Type);
    }
}

static PyObject *bits_acpi_eval(PyObject *self, PyObject *args)
{
    char *pathname;
    ACPI_BUFFER results = { .Length = ACPI_ALLOCATE_BUFFER, .Pointer = NULL };
    PyObject *ret;

    if (!PyArg_ParseTuple(args, "s", &pathname))
        return NULL;

    if (acpica_init() != GRUB_ERR_NONE)
        return PyErr_Format(PyExc_RuntimeError, "ACPICA module failed to initialize.");

    if (ACPI_FAILURE(AcpiEvaluateObject(NULL, pathname, NULL, &results)))
        return Py_BuildValue("");

    ret = acpi_object_to_python(results.Pointer);

    ACPI_FREE(results.Pointer);
    return Py_BuildValue("O", ret);
}

static PyObject *bits_acpi_get_table(PyObject *self, PyObject *args)
{
    char *signature;
    U32 instance = 0;
    ACPI_TABLE_HEADER *table_header;

    if (!PyArg_ParseTuple(args, "s|I", &signature, instance))
        return NULL;

    if (acpica_init() != GRUB_ERR_NONE)
        return PyErr_Format(PyExc_RuntimeError, "ACPICA module failed to initialize.");

    if (ACPI_FAILURE(AcpiGetTable(signature, instance, &table_header)))
        return Py_BuildValue("");

    return Py_BuildValue("s#", table_header, table_header->Length);
}

struct find_processor_context {
    bool init_cpu;
    U32 caps;
    PyObject *cpupath_list;
};

static grub_err_t call_osc(ACPI_HANDLE cpu_handle, U32 caps)
{
    grub_err_t ret;
    ACPI_STATUS Status;
    ACPI_BUFFER Results = { .Length = ACPI_ALLOCATE_BUFFER, .Pointer = NULL };
    ACPI_OBJECT_LIST Params;
    ACPI_OBJECT Obj[4];
    U32 osc_buffer[2];
    U8 OSC_UUID[16] = {0x16, 0xA6, 0x77, 0x40, 0x0C, 0x29, 0xBE, 0x47, 0x9E, 0xBD, 0xD8, 0x70, 0x58, 0x71, 0x39, 0x53};

    /* Initialize the parameter list */
    Params.Count = 4;
    Params.Pointer = Obj;

    /* Initialize the parameter objects */
    // Intel-specific UUID
    Obj[0].Type = ACPI_TYPE_BUFFER;
    Obj[0].Buffer.Length = 16;
    Obj[0].Buffer.Pointer = OSC_UUID;

    // Revision ID
    Obj[1].Type = ACPI_TYPE_INTEGER;
    Obj[1].Integer.Value = 1;

    // Count of DWORDS
    Obj[2].Type = ACPI_TYPE_INTEGER;
    Obj[2].Integer.Value = 2;

    // Capabilities buffer
    Obj[3].Type = ACPI_TYPE_BUFFER;
    Obj[3].Buffer.Length = 8;
    Obj[3].Buffer.Pointer = (U8 *)osc_buffer;

    osc_buffer[0] = 0;
    osc_buffer[1] = caps; // Capabilities DWORD

    Status = AcpiEvaluateObject(cpu_handle, "_OSC", &Params, &Results);
    if (Status == AE_OK)
        ret = GRUB_ERR_NONE;
    else if (Status == AE_NOT_FOUND)
        ret = grub_error(GRUB_ERR_TEST_FAILURE, "false");
    else
        ret = grub_error(GRUB_ERR_IO, "Evaluating _OSC failed (0x%x %s)\n", Status, AcpiFormatException(Status));

    ACPI_FREE(Results.Pointer);

    return ret;
}

static grub_err_t call_pdc(ACPI_HANDLE cpu_handle, U32 caps)
{
    ACPI_STATUS Status;
    ACPI_OBJECT_LIST Params;
    ACPI_OBJECT Obj;
    U32 pdc_buffer[3];

    /* Initialize the parameter list */
    Params.Count = 1;
    Params.Pointer = &Obj;

    /* Initialize the parameter objects */
    Obj.Type = ACPI_TYPE_BUFFER;
    Obj.Buffer.Length = 12;
    Obj.Buffer.Pointer = (UINT8 *)pdc_buffer;

    /* Initialize the DWORD array */
    pdc_buffer[0] = 1; // Revision
    pdc_buffer[1] = 1; // Count
    pdc_buffer[2] = caps; // Capabilities DWORD

    Status = AcpiEvaluateObject(cpu_handle, "_PDC", &Params, NULL);
    if (Status == AE_OK)
        return GRUB_ERR_NONE;
    else if (Status == AE_NOT_FOUND)
        return grub_error(GRUB_ERR_TEST_FAILURE, "false");
    else
        return grub_error(GRUB_ERR_IO, "Evaluating _PDC failed (0x%x %s)\n", Status, AcpiFormatException(Status));
}

static ACPI_STATUS find_processor(ACPI_HANDLE ObjHandle, UINT32 NestingLevel ACPI_UNUSED_VAR, void *Context, void **ReturnValue ACPI_UNUSED_VAR)
{
    struct find_processor_context *fpc = Context;
    ACPI_BUFFER Path = { .Length = ACPI_ALLOCATE_BUFFER, .Pointer = NULL };
    PyObject *cpupath;

    if (!IsEnabledProcessor(ObjHandle))
        goto out;

    if (ACPI_FAILURE(AcpiGetName(ObjHandle, ACPI_FULL_PATHNAME, &Path))) {
        grub_printf("Couldn't get object name\n");
        goto out;
    }

    if (fpc->init_cpu)
        if (call_osc(ObjHandle, fpc->caps) != GRUB_ERR_NONE)
            call_pdc(ObjHandle, fpc->caps);

    cpupath = Py_BuildValue("s", Path.Pointer);
    PyList_Append(fpc->cpupath_list, cpupath);
    Py_XDECREF(cpupath);

out:
    ACPI_FREE(Path.Pointer);
    return AE_OK;
}

static PyObject *bits_acpi_cpupaths(PyObject *self, PyObject *args)
{
    struct find_processor_context fpc = { .init_cpu = false, .caps = 0xfbf, .cpupath_list = NULL };

    if (acpica_init() != GRUB_ERR_NONE)
        return PyErr_Format(PyExc_RuntimeError, "ACPICA module failed to initialize.");

    // Before getting any input parameters, change the value of the capabilities DWORD
    // to match the value used in prior ACPI CPU initialization
    if (acpica_cpus_initialized)
        fpc.caps = acpica_cpus_init_caps;

    if (!PyArg_ParseTuple(args, "|I", &fpc.caps))
        return NULL;

    if (acpica_cpus_initialized) {
        if (fpc.caps != acpica_cpus_init_caps)
            return PyErr_Format(PyExc_ValueError, "Attempt to change current Capabilities DWORD from 0x%x to 0x%x; changing capabilities requires ACPI shutdown (acpi_terminate) and restart.",
                                acpica_cpus_init_caps, fpc.caps);
    } else {
        acpica_cpus_initialized = true;
        acpica_cpus_init_caps = fpc.caps;
        fpc.init_cpu = true;
    }

    fpc.cpupath_list = PyList_New(0);

    AcpiWalkNamespace(ACPI_TYPE_PROCESSOR, ACPI_ROOT_OBJECT, ACPI_UINT32_MAX, find_processor, NULL, &fpc, NULL);

    return fpc.cpupath_list;
}

struct find_object_context {

    char *needle;
    PyObject *objpath_list;
};

static ACPI_STATUS find_object(ACPI_HANDLE ObjHandle, UINT32 NestingLevel ACPI_UNUSED_VAR, void *Context, void **ReturnValue ACPI_UNUSED_VAR)
{
    struct find_object_context *foc = Context;
    ACPI_BUFFER Path = { .Length = ACPI_ALLOCATE_BUFFER, .Pointer = NULL };
    PyObject *objpath;

    if (ACPI_FAILURE(AcpiGetName(ObjHandle, ACPI_FULL_PATHNAME, &Path))) {
        grub_printf("Couldn't get object name\n");
        goto out;
    }

    if (grub_strstr(Path.Pointer, foc->needle) == NULL)
        goto out;

    objpath = Py_BuildValue("s", Path.Pointer);
    PyList_Append(foc->objpath_list, objpath);
    Py_XDECREF(objpath);

out:
    ACPI_FREE(Path.Pointer);
    return AE_OK;
}

static PyObject *bits_acpi_objpaths(PyObject *self, PyObject *args)
{
    struct find_object_context foc = { .objpath_list = NULL };

    if (acpica_init() != GRUB_ERR_NONE)
        return PyErr_Format(PyExc_RuntimeError, "ACPICA module failed to initialize.");

    if (!PyArg_ParseTuple(args, "s", &foc.needle))
        return NULL;

    foc.objpath_list = PyList_New(0);

    AcpiWalkNamespace(ACPI_TYPE_ANY, ACPI_ROOT_OBJECT, ACPI_UINT32_MAX, find_object, NULL, &foc, NULL);

    return foc.objpath_list;
}

static PyMethodDef bitsMethods[] = {
    {"acpi_cpupaths",  bits_acpi_cpupaths, METH_VARARGS, "acpi_cpupaths([capabilities]) -> list of cpu namepaths"},
    {"acpi_eval",  bits_acpi_eval, METH_VARARGS, "acpi_eval(\"\\PATH._TO_.EVAL\") -> result"},
    {"acpi_get_table",  bits_acpi_get_table, METH_VARARGS, "acpi_get_table(signature[, instance]) -> str"},
    {"acpi_objpaths",  bits_acpi_objpaths, METH_VARARGS, "acpi_objpaths(\"objectname\") -> list of obj namepaths"},
    {"cpuid", bits_cpuid, METH_VARARGS, "cpuid(apicid, eax[, ecx]) -> eax, ebx, ecx, edx"},
    {"cpus",  bits_cpus, METH_NOARGS, "cpus() -> list of APIC IDs"},
    {"pci_read", (PyCFunction)bits_pci_read, METH_KEYWORDS, "pci_read(bus, device, function, register[, bytes=BYTES]) -> PCIe register value"},
    {"pci_write", (PyCFunction)bits_pci_write, METH_KEYWORDS, "pci_write(bus, device, function, register, value[, bytes=BYTES]"},
    {"pcie_read", (PyCFunction)bits_pcie_read, METH_KEYWORDS, "pcie_read(bus, device, function, register[, bytes=BYTES][, memaddr=MEMADDR]) -> PCIe register value"},
    {"pcie_write", (PyCFunction)bits_pcie_write, METH_KEYWORDS, "pcie_write(bus, device, function, register, value[, bytes=BYTES][, memaddr=MEMADDR])"},
    {"pcie_get_base", bits_pcie_get_base, METH_NOARGS, "pcie_get_base() -> memory address" },
    {"pcie_set_base", bits_pcie_set_base, METH_VARARGS, "pcie_set_base(memaddr)" },
    {"rdmsr",  bits_rdmsr, METH_VARARGS, "rdmsr(apicid, msr) -> long (None if GPF)"},
    {"wrmsr",  bits_wrmsr, METH_VARARGS, "wrmsr(apicid, msr, value) -> bool (1 if GPF, else 0)"},
    {NULL, NULL, 0, NULL}        /* Sentinel */
};

PyMODINIT_FUNC init_bits(void)
{
    (void) Py_InitModule("_bits", bitsMethods);
}
